import shutil

import pytest

from tests.helpers.api import ConnectorApi
from tests.helpers.base_helper import prepare_ns_with_items


CONNECTIONS = []


@pytest.fixture
def db(request):
    if request.config.getoption('--mode') != "builtin":
        pytest.skip("tests only for builtin mode")

    def _inner(**kwargs):
        db = ConnectorApi("builtin://tmp/test_db", **kwargs)
        CONNECTIONS.append(db)
        return db

    yield _inner


@pytest.fixture(autouse=True)
def clear_storage(db):
    yield
    for conn in CONNECTIONS:
        conn.close()
    CONNECTIONS.clear()
    shutil.rmtree("tmp/", ignore_errors=True)


class TestBuiltinOptions:
    def test_builtin_options(self, db):
        db = db(max_replication_updates_size=1, allocator_cache_limit=1, allocator_cache_part=1)
        prepare_ns_with_items(db)

    def test_builtin_options_zero_values(self, db):
        db = db(max_replication_updates_size=0, allocator_cache_limit=0, allocator_cache_part=0)
        prepare_ns_with_items(db)

    def test_builtin_options_negative_values(self, db):
        db = db(max_replication_updates_size=-1000, allocator_cache_limit=-1000, allocator_cache_part=-1000)
        prepare_ns_with_items(db)



from datetime import timedelta
from typing import Dict, List

from pyreindexer.query import Query
from raiser_mixin import RaiserTx


class Transaction(RaiserTx):
    """An object representing the context of a Reindexer transaction

    #### Attributes:
        api (module): An API module for Reindexer calls
        transaction_wrapper_ptr (int): A memory pointer to Reindexer transaction object
        err_code (int): The API error code
        err_msg (string): The API error message

    """

    def __init__(self, rx, transaction_wrapper_ptr: int):
        """Constructs a new Reindexer transaction object

        #### Arguments:
            api (module): An API module for Reindexer calls
            transaction_wrapper_ptr (int): A memory pointer to Reindexer transaction object

        """

        self.rx = rx
        self.api = rx.api
        self.transaction_wrapper_ptr: int = transaction_wrapper_ptr
        self.err_code: int = 0
        self.err_msg: str = ""

    def __del__(self):
        """Rollbacks a transaction if it was not previously stopped

        """

        if self.transaction_wrapper_ptr > 0 and self.rx.rx > 0:
            self.rollback(timedelta(milliseconds=0))

    def __finalize(self):
        with self.rx._tx_lock:
            self.rx._tx_ptrs.remove(self.transaction_wrapper_ptr)
        self.transaction_wrapper_ptr = 0

    @RaiserTx.raise_if_error
    def insert(self, item_def: Dict, precepts: List[str] = None) -> None:
        """Inserts an item with its precepts to the transaction
            Warning: the timeout set when the transaction was created is used

        #### Arguments:
            item_def (dict): A dictionary of item definition
            precepts (:obj:`list` of :obj:`str`): A dictionary of index definition

        #### Raises:
            TransactionError: Raises with an error message of API return if Transaction is over
            ApiError: Raises with an error message of API return on non-zero error code

        """

        precepts = [] if precepts is None else precepts
        self.err_code, self.err_msg = self.api.item_insert_transaction(self.transaction_wrapper_ptr,
                                                                       item_def, precepts)

    @RaiserTx.raise_if_error
    def update(self, item_def: Dict, precepts: List[str] = None) -> None:
        """Updates an item with its precepts to the transaction
            Warning: the timeout set when the transaction was created is used

        #### Arguments:
            item_def (dict): A dictionary of item definition
            precepts (:obj:`list` of :obj:`str`): A dictionary of index definition

        #### Raises:
            TransactionError: Raises with an error message of API return if Transaction is over
            ApiError: Raises with an error message of API return on non-zero error code

        """

        precepts = [] if precepts is None else precepts
        self.err_code, self.err_msg = self.api.item_update_transaction(self.transaction_wrapper_ptr,
                                                                       item_def, precepts)

    @RaiserTx.raise_if_error
    def update_query(self, query: Query) -> None:
        """Updates items with the transaction
            Read-committed isolation is available for read operations.
            Changes made in active transaction is invisible to current and another transactions.

        #### Arguments:
            query (:obj:`Query`): A query object to modify

        #### Raises:
            TransactionError: Raises with an error message of API return if Transaction is over
            ApiError: Raises with an error message of API return on non-zero error code

        """

        self.err_code, self.err_msg = self.api.modify_transaction(self.transaction_wrapper_ptr,
                                                                  query.query_wrapper_ptr)

    @RaiserTx.raise_if_error
    def upsert(self, item_def: Dict, precepts: List[str] = None) -> None:
        """Updates an item with its precepts to the transaction. Creates the item if it does not exist
            Warning: the timeout set when the transaction was created is used

        #### Arguments:
            item_def (dict): A dictionary of item definition
            precepts (:obj:`list` of :obj:`str`): A dictionary of index definition

        #### Raises:
            TransactionError: Raises with an error message of API return if Transaction is over
            ApiError: Raises with an error message of API return on non-zero error code

        """

        precepts = [] if precepts is None else precepts
        self.err_code, self.err_msg = self.api.item_upsert_transaction(self.transaction_wrapper_ptr,
                                                                       item_def, precepts)

    @RaiserTx.raise_if_error
    def delete(self, item_def: Dict) -> None:
        """Deletes an item from the transaction
            Warning: the timeout set when the transaction was created is used

        #### Arguments:
            item_def (dict): A dictionary of item definition

        #### Raises:
            TransactionError: Raises with an error message of API return if Transaction is over
            ApiError: Raises with an error message of API return on non-zero error code

        """

        self.err_code, self.err_msg = self.api.item_delete_transaction(self.transaction_wrapper_ptr,
                                                                       item_def)

    @RaiserTx.raise_if_error
    def delete_query(self, query: Query):
        """Deletes items with the transaction
            Read-committed isolation is available for read operations.
            Changes made in active transaction is invisible to current and another transactions.

        #### Arguments:
            query (:obj:`Query`): A query object to modify

        #### Raises:
            TransactionError: Raises with an error message of API return if Transaction is over
            ApiError: Raises with an error message of API return on non-zero error code

        """

        self.err_code, self.err_msg = self.api.delete_transaction(self.transaction_wrapper_ptr,
                                                                  query.query_wrapper_ptr)

    @RaiserTx.raise_if_error
    def commit(self, timeout: timedelta = timedelta(milliseconds=0)) -> None:
        """Applies changes

        #### Arguments:
            timeout (`datetime.timedelta`): Optional timeout for performing a server-side operation.
                Minimum is 1 millisecond; if set to a lower value, it corresponds to disabling the timeout.
                A value of 0 disables the timeout (default value)

        #### Raises:
            TransactionError: Raises with an error message of API return if Transaction is over
            ApiError: Raises with an error message of API return on non-zero error code

        """

        milliseconds: int = int(timeout / timedelta(milliseconds=1))
        self.err_code, self.err_msg, _ = self.api.commit_transaction(self.transaction_wrapper_ptr,
                                                                     milliseconds)
        self.__finalize()

    @RaiserTx.raise_if_error
    def commit_with_count(self, timeout: timedelta = timedelta(milliseconds=0)) -> int:
        """Applies changes and return the number of count of changed items

        #### Arguments:
            timeout (`datetime.timedelta`): Optional timeout for performing a server-side operation.
                Minimum is 1 millisecond; if set to a lower value, it corresponds to disabling the timeout.
                A value of 0 disables the timeout (default value)

        #### Raises:
            TransactionError: Raises with an error message of API return if Transaction is over
            ApiError: Raises with an error message of API return on non-zero error code

        """

        milliseconds: int = int(timeout / timedelta(milliseconds=1))
        self.err_code, self.err_msg, count = self.api.commit_transaction(self.transaction_wrapper_ptr,
                                                                         milliseconds)
        self.__finalize()
        return count

    @RaiserTx.raise_if_error
    def rollback(self, timeout: timedelta = timedelta(milliseconds=0)) -> None:
        """Rollbacks changes

        #### Arguments:
            timeout (`datetime.timedelta`): Optional timeout for performing a server-side operation.
                Minimum is 1 millisecond; if set to a lower value, it corresponds to disabling the timeout.
                A value of 0 disables the timeout (default value)

        #### Raises:
            TransactionError: Raises with an error message of API return if Transaction is over
            ApiError: Raises with an error message of API return on non-zero error code

        """

        milliseconds: int = int(timeout / timedelta(milliseconds=1))
        self.err_code, self.err_msg = self.api.rollback_transaction(self.transaction_wrapper_ptr,
                                                                    milliseconds)
        self.__finalize()



index_definition = {
    "name": "id",
    "field_type": "int",
    "index_type": "hash",
    "is_pk": True,
    "is_array": False,
    "is_dense": False,
    "is_sparse": False,
    'is_no_column': False,
    "collate_mode": "none",
    "sort_order_letters": "",
    "config": {},
    "expire_after": 0,
    "json_paths": ["id"]
}

updated_index_definition = {
    "name": "id",
    "field_type": "int64",
    "index_type": "hash",
    "is_pk": True,
    "is_array": False,
    "is_dense": False,
    "is_sparse": False,
    'is_no_column': False,
    "collate_mode": "none",
    "sort_order_letters": "",
    "config": {},
    "expire_after": 0,
    "json_paths": ["id_new"]
}

composite_index_definition = {
    "name": "comp_idx",
    "field_type": "composite",
    "index_type": "hash",
    "is_pk": False,
    "is_array": False,
    "is_dense": False,
    "is_sparse": False,
    "json_paths": ["id", "val"]
}

vector_index_bf = {
    "name": "vec",
    "field_type": "float_vector",
    "index_type": "vec_bf",
    "is_pk": False,
    "is_array": False,
    "is_dense": False,
    "is_sparse": False,
    'is_no_column': False,
    "collate_mode": "none",
    "sort_order_letters": "",
    "expire_after": 0,
    "config": {"dimension": 4, "metric": "l2", "start_size": 1000},
    "json_paths": ["vec"]
}

vector_index_hnsw = {
    "name": "vec",
    "field_type": "float_vector",
    "index_type": "hnsw",
    "is_pk": False,
    "is_array": False,
    "is_dense": False,
    "is_sparse": False,
    'is_no_column': False,
    "collate_mode": "none",
    "sort_order_letters": "",
    "expire_after": 0,
    "config": {"dimension": 4, "metric": "inner_product", "start_size": 1000,
               "ef_construction": 20, "m": 4, "multithreading": 1},
    "json_paths": ["vec"]
}

vector_index_ivf = {
    "name": "vec",
    "field_type": "float_vector",
    "index_type": "ivf",
    "is_pk": False,
    "is_array": False,
    "is_dense": False,
    "is_sparse": False,
    'is_no_column': False,
    "collate_mode": "none",
    "sort_order_letters": "",
    "expire_after": 0,
    "config": {"dimension": 4, "metric": "cosine", "centroids_count": 10},
    "json_paths": ["vec"]
}

special_namespaces = [{"name": "#namespaces"},
                      {"name": "#memstats"},
                      {"name": "#perfstats"},
                      {"name": "#config"},
                      {"name": "#queriesperfstats"},
                      {"name": "#activitystats"},
                      {"name": "#clientsstats"}]

special_namespaces_cluster = [{"name": "#namespaces"},
                              {"name": "#memstats"},
                              {"name": "#perfstats"},
                              {"name": "#config"},
                              {"name": "#queriesperfstats"},
                              {"name": "#activitystats"},
                              {"name": "#clientsstats"},
                              {"name": "#replicationstats"}]

item_definition = {"id": 100, "val": "testval"}

AGGREGATE_FUNCTIONS_MATH = [(lambda x: max(x), "aggregate_max"),
                            (lambda x: min(x), "aggregate_min"),
                            (lambda x: sum(x), "aggregate_sum"),
                            (lambda x: sum(x) / len(x), "aggregate_avg")]



from datetime import timedelta
from typing import Final

from pyreindexer import RxConnector
from query import Query
from tests.helpers.log_helper import log_api


def make_request_and_response_log(method_description, call_msg, res=None) -> str:
    return f"{method_description}\n\t[Call] => {call_msg}\n\t[Return] <= {res}"


def api_method(func):
    def wrapped(self, *args, **kwargs):
        try:
            method_description = func.__doc__.split("\n")[0]
        except AttributeError:
            raise RuntimeError("Api Method doesn't have a 'docstring' description")

        args_str = ", ".join(repr(a) for a in args)
        kwargs_str = ", ".join(f"{k}={v}" for k, v in kwargs.items())
        call_msg = f"args ({args_str}), kwargs ({kwargs_str})"

        r = func(self, *args, **kwargs)
        log = make_request_and_response_log(method_description, call_msg, r)
        log_api.info(log)
        return r

    return wrapped


class ConnectorApi(RxConnector):

    def __init__(self, dsn, **kwargs):
        super().__init__(dsn, **kwargs)
        self.namespace = NamespaceApiMethods(self)
        self.index = IndexApiMethods(self)
        self.item = ItemApiMethods(self)
        self.query = QueryApiMethods(self)
        self.meta = MetaApiMethods(self)
        self.tx = TransactionApiMethods(self)

    default_timeout: Final[timedelta] = timedelta(milliseconds=0)


class NamespaceApiMethods:
    def __init__(self, api):
        self.api = api

    @api_method
    def open(self, ns_name, timeout: timedelta = ConnectorApi.default_timeout):
        """ Open namespace """
        return self.api.namespace_open(ns_name, timeout)

    @api_method
    def close(self, ns_name, timeout: timedelta = ConnectorApi.default_timeout):
        """ Close namespace """
        return self.api.namespace_close(ns_name, timeout)

    @api_method
    def drop(self, ns_name, timeout: timedelta = ConnectorApi.default_timeout):
        """ Drop namespace """
        return self.api.namespace_drop(ns_name, timeout)

    @api_method
    def enumerate(self, enum_not_opened=False, timeout: timedelta = ConnectorApi.default_timeout):
        """ Get namespaces list """
        return self.api.namespaces_enum(enum_not_opened, timeout)


class IndexApiMethods:
    def __init__(self, api):
        self.api = api

    @api_method
    def create(self, ns_name, index, timeout: timedelta = ConnectorApi.default_timeout):
        """ Add index """
        return self.api.index_add(ns_name, index, timeout)

    @api_method
    def update(self, ns_name, index, timeout: timedelta = ConnectorApi.default_timeout):
        """ Update index """
        return self.api.index_update(ns_name, index, timeout)

    @api_method
    def drop(self, ns_name, index_name, timeout: timedelta = ConnectorApi.default_timeout):
        """ Drop index """
        return self.api.index_drop(ns_name, index_name, timeout)


class ItemApiMethods:
    def __init__(self, api):
        self.api = api

    @api_method
    def insert(self, ns_name, item, precepts=None, timeout: timedelta = ConnectorApi.default_timeout):
        """ Insert item """
        return self.api.item_insert(ns_name, item, precepts, timeout)

    @api_method
    def upsert(self, ns_name, item, precepts=None, timeout: timedelta = ConnectorApi.default_timeout):
        """ Upsert item """
        return self.api.item_upsert(ns_name, item, precepts, timeout)

    @api_method
    def update(self, ns_name, item, precepts=None, timeout: timedelta = ConnectorApi.default_timeout):
        """ Update item """
        return self.api.item_update(ns_name, item, precepts, timeout)

    @api_method
    def delete(self, ns_name, item, timeout: timedelta = ConnectorApi.default_timeout):
        """ Delete item """
        return self.api.item_delete(ns_name, item, timeout)


class QueryApiMethods:
    def __init__(self, api):
        self.api = api

    @api_method
    def sql(self, q, timeout: timedelta = ConnectorApi.default_timeout):
        """ Execute SQL query """
        return self.api.exec_sql(q, timeout)

    @api_method
    def new(self, ns_name) -> Query:
        """ Create a new query """
        return self.api.new_query(ns_name)


class MetaApiMethods:
    def __init__(self, api):
        self.api = api

    @api_method
    def put(self, ns_name, key, value, timeout: timedelta = ConnectorApi.default_timeout):
        """ Put meta with key and value """
        return self.api.meta_put(ns_name, key, value, timeout)

    @api_method
    def get(self, ns_name, key, timeout: timedelta = ConnectorApi.default_timeout):
        """ Get meta by key """
        return self.api.meta_get(ns_name, key, timeout)

    @api_method
    def enumerate(self, ns_name, timeout: timedelta = ConnectorApi.default_timeout):
        """ Get meta keys list """
        return self.api.meta_enum(ns_name, timeout)

    @api_method
    def delete(self, ns_name, key, timeout: timedelta = ConnectorApi.default_timeout):
        """ Delete meta by key """
        return self.api.meta_delete(ns_name, key, timeout)


class TransactionApiMethods:
    def __init__(self, api):
        self.api = api

    class _TransactionApi:
        def __init__(self, tx):
            self.tx = tx

        @api_method
        def commit(self, timeout: timedelta = ConnectorApi.default_timeout):
            """ Commit the transaction """
            return self.tx.commit(timeout)

        @api_method
        def commit_with_count(self, timeout: timedelta = ConnectorApi.default_timeout):
            """ Commit the transaction and return the number of changed items """
            return self.tx.commit_with_count(timeout)

        @api_method
        def rollback(self, timeout: timedelta = ConnectorApi.default_timeout):
            """ Rollback the transaction """
            return self.tx.rollback(timeout)

        @api_method
        def insert_item(self, item, precepts=None):
            """ Insert item into transaction """
            return self.tx.insert(item, precepts)

        @api_method
        def upsert_item(self, item, precepts=None):
            """ Upsert item into transaction """
            return self.tx.upsert(item, precepts)

        @api_method
        def update_item(self, item, precepts=None):
            """ Update item into transaction """
            return self.tx.update(item, precepts)

        @api_method
        def delete_item(self, item):
            """ Delete item from transaction """
            return self.tx.delete(item)

        @api_method
        def update_query(self, query):
            """ Call update query in transaction """
            return self.tx.update_query(query)

        @api_method
        def delete_query(self, query):
            """ Call delete query in transaction """
            return self.tx.delete_query(query)

    @api_method
    def begin(self, ns_name, timeout: timedelta = ConnectorApi.default_timeout) -> "_TransactionApi":
        """ Begin new transaction """
        tx = self.api.new_transaction(ns_name, timeout)
        return self._TransactionApi(tx)



from __future__ import annotations

import threading
from datetime import timedelta
from typing import Dict, List

from pyreindexer.query import Query
from pyreindexer.query_results import QueryResults
from pyreindexer.raiser_mixin import RaiserRx
from pyreindexer.transaction import Transaction


class RxConnector(RaiserRx):
    """RxConnector provides a binding to Reindexer upon two shared libraries (hereinafter - APIs): 'rawpyreindexerb.so'
        and 'rawpyreindexerc.so'. The first one is aimed at builtin usage. That API embeds Reindexer, so it could
        be used right in-place as is. The second one acts as a lightweight client which establishes a connection to
        Reindexer server via RPC. The APIs interfaces are completely the same.

    #### Arguments:
            dsn (string): The connection string which contains a protocol
                Examples: 'builtin:///tmp/pyrx', 'cproto://127.0.0.1:6534/pyrx'

            cproto options:
                 fetch_amount (int): The number of items that will be fetched by one operation
                 reconnect_attempts (int): Number of reconnection attempts when connection is lost
                 net_timeout (`datetime.timedelta`): Connection and database login timeout value [milliseconds]
                 enable_compression (bool): Flag enable/disable traffic compression
                 start_special_thread (bool): Determines whether to request a special thread of execution
                    on the server for this connection
                 client_name (string): Proper name of the application (as a client for Reindexer-server)
                 sync_rxcoro_count (int): Client concurrency per connection [1..10'000], default 10

            built-in options:
                max_replication_updates_size (int): Max pended replication updates size in bytes
                allocator_cache_limit (int): Recommended maximum free cache size of tcmalloc memory allocator in bytes
                allocator_cache_part (float): Recommended maximum free cache size of tcmalloc memory allocator in
                    relation to total Reindexer allocated memory size, in units

    #### Attributes:
        api (module): An API module loaded dynamically for Reindexer calls
        rx (int): A memory pointer to the Reindexer instance
        err_code (int): The API error code
        err_msg (string): The API error message

    """

    def __init__(self, dsn: str, *,
                 # cproto options
                 fetch_amount: int = 1000,
                 reconnect_attempts: int = 0,
                 net_timeout: timedelta = timedelta(milliseconds=0),
                 enable_compression: bool = False,
                 start_special_thread: bool = False,
                 client_name: str = 'pyreindexer',
                 sync_rxcoro_count: int = 10,
                 # builtin options
                 max_replication_updates_size: int = 1024 * 1024 * 1024,
                 allocator_cache_limit: int = -1,
                 allocator_cache_part: float = -1.0):
        """Constructs a new connector object.
        Initializes an error code and a Reindexer instance descriptor to zero

        #### Arguments:
            dsn (string): The connection string which contains a protocol
                Examples: 'builtin:///tmp/pyrx', 'cproto://127.0.0.1:6534/pyrx'

            cproto options:
                 fetch_amount (int): The number of items that will be fetched by one operation (must be > 0)
                 reconnect_attempts (int): Number of reconnection attempts when connection is lost
                 net_timeout (`datetime.timedelta`): Connection and database login timeout value [milliseconds]
                 enable_compression (bool): Flag enable/disable traffic compression
                 start_special_thread (bool): Determines whether to request a special thread of execution
                    on the server for this connection
                 client_name (string): Proper name of the application (as a client for Reindexer-server)
                 sync_rxcoro_count (int): Client concurrency per connection [1..10'000], default 10

            built-in options:
                max_replication_updates_size (int): Max pended replication updates size in bytes
                allocator_cache_limit (int): Recommended maximum free cache size of tcmalloc memory allocator in bytes
                allocator_cache_part (float): Recommended maximum free cache size of tcmalloc memory allocator in
                    relation to total Reindexer allocated memory size, in units

        #### Raises:
            ValueError: Raises with an error message when input value is invalid

        """

        self.err_code: int = 0
        self.err_msg: str = ''
        self.rx: int = 0

        if fetch_amount <= 0:
            raise ValueError("'fetch_amount' must be greater than zero")

        self._api_import(dsn)
        milliseconds: int = int(net_timeout / timedelta(milliseconds=1))
        self.rx = self.api.init(fetch_amount, reconnect_attempts, milliseconds, enable_compression,
                                start_special_thread, client_name, sync_rxcoro_count,
                                max_replication_updates_size, allocator_cache_limit, allocator_cache_part)
        self._api_connect(dsn, net_timeout)
        self._tx_ptrs = set()
        self._tx_lock = threading.Lock()
        self._query_ptrs = set()
        self._query_lock = threading.Lock()

    def __del__(self):
        """Closes an API instance on a connector object deletion if the API is initialized

        """

        if self.rx > 0:
            with self._tx_lock:
                tx_ptrs_list = list(self._tx_ptrs)
                self._tx_ptrs.clear()
            with self._query_lock:
                query_ptrs_list = list(self._query_ptrs)
                self._query_ptrs.clear()

            for tx_ptr in tx_ptrs_list:
                self.api.rollback_transaction(tx_ptr, 0)
            for q_ptr in query_ptrs_list:
                self.api.destroy_query(q_ptr)

            self._api_close()

    def _api_import(self, dsn: str) -> None:
        """Imports an API dynamically depending on protocol specified in dsn

        #### Arguments:
            dsn (string): The connection string which contains a protocol

        #### Raises:
            ConnectionError: Raises an exception if a connection protocol is unrecognized

        """

        if dsn.startswith('builtin://'):
            self.api = __import__('rawpyreindexerb')
        elif dsn.startswith('cproto://'):
            self.api = __import__('rawpyreindexerc')
        else:
            raise ConnectionError(f"Unknown Reindexer connection protocol for dsn: {dsn}")

    def _api_connect(self, dsn: str, timeout: timedelta) -> None:
        """Connects to a database specified in dsn. Obtains a pointer to Reindexer instance

        #### Arguments:
            dsn (string): The connection string which contains a protocol
            timeout (`datetime.timedelta`): Optional timeout for performing a server-side operation.
                Minimum is 1 millisecond; if set to a lower value, it corresponds to disabling the timeout.
                A value of 0 disables the timeout (default value)

        #### Raises:
            ConnectionError: Raises with an error message when Reindexer instance is not initialized yet
            ApiError: Raises with an error message of API return on non-zero error code

        """

        self.raise_on_not_init()
        milliseconds: int = int(timeout / timedelta(milliseconds=1))
        self.err_code, self.err_msg = self.api.connect(self.rx, dsn, milliseconds)
        self.raise_on_error()

    def _api_close(self) -> None:
        """Destructs Reindexer instance correctly and resets memory pointer

        #### Raises:
            ConnectionError: Raises with an error message when Reindexer instance is not initialized yet

        """

        self.raise_on_not_init()
        self.api.destroy(self.rx)
        self.rx = 0

    def close(self) -> None:
        """Closes the API instance and frees Reindexer resources

        #### Raises:
            ConnectionError: Raises with an error message when Reindexer instance is not initialized yet

        """

        self._api_close()

    @RaiserRx.raise_if_error
    def namespace_open(self, namespace: str, timeout: timedelta = timedelta(milliseconds=0)) -> None:
        """Opens a namespace specified or creates a namespace if it does not exist

        #### Arguments:
            namespace (string): The name of the namespace
            timeout (`datetime.timedelta`): Optional timeout for performing a server-side operation.
                Minimum is 1 millisecond; if set to a lower value, it corresponds to disabling the timeout.
                A value of 0 disables the timeout (default value)

        #### Raises:
            ConnectionError: Raises with an error message when Reindexer instance is not initialized yet
            ApiError: Raises with an error message of API return on non-zero error code

        """

        milliseconds: int = int(timeout / timedelta(milliseconds=1))
        self.err_code, self.err_msg = self.api.namespace_open(self.rx, namespace, milliseconds)

    @RaiserRx.raise_if_error
    def namespace_close(self, namespace: str, timeout: timedelta = timedelta(milliseconds=0)) -> None:
        """Closes the specified namespace

        #### Arguments:
            namespace (string): The name of the namespace
            timeout (`datetime.timedelta`): Optional timeout for performing a server-side operation.
                Minimum is 1 millisecond; if set to a lower value, it corresponds to disabling the timeout.
                A value of 0 disables the timeout (default value)

        #### Raises:
            ConnectionError: Raises with an error message when Reindexer instance is not initialized yet
            ApiError: Raises with an error message of API return on non-zero error code

        """

        milliseconds: int = int(timeout / timedelta(milliseconds=1))
        self.err_code, self.err_msg = self.api.namespace_close(self.rx, namespace, milliseconds)

    @RaiserRx.raise_if_error
    def namespace_drop(self, namespace: str, timeout: timedelta = timedelta(milliseconds=0)) -> None:
        """Drops the specified namespace

        #### Arguments:
            namespace (string): The name of the namespace
            timeout (`datetime.timedelta`): Optional timeout for performing a server-side operation.
                Minimum is 1 millisecond; if set to a lower value, it corresponds to disabling the timeout.
                A value of 0 disables the timeout (default value)

        #### Raises:
            Exception: Raises with an error message when Reindexer instance is not initialized yet
            Exception: Raises with an error message of API return on non-zero error code

        """

        milliseconds: int = int(timeout / timedelta(milliseconds=1))
        self.err_code, self.err_msg = self.api.namespace_drop(self.rx, namespace, milliseconds)

    @RaiserRx.raise_if_error
    def namespaces_enum(self, enum_not_opened: bool = False,
                        timeout: timedelta = timedelta(milliseconds=0)) -> List[Dict[str, str]]:
        """Gets a list of namespaces available

        #### Arguments:
            enum_not_opened (bool, optional): An enumeration mode flag. If it is
                set then closed namespaces are in result list too. Defaults to False
            timeout (`datetime.timedelta`): Optional timeout for performing a server-side operation.
                Minimum is 1 millisecond; if set to a lower value, it corresponds to disabling the timeout.
                A value of 0 disables the timeout (default value)

        #### Returns:
            (:obj:`list` of :obj:`dict`): A list of dictionaries which describe each namespace

        #### Raises:
            ConnectionError: Raises with an error message when Reindexer instance is not initialized yet
            ApiError: Raises with an error message of API return on non-zero error code

        """

        milliseconds: int = int(timeout / timedelta(milliseconds=1))
        self.err_code, self.err_msg, res = self.api.namespaces_enum(self.rx, enum_not_opened, milliseconds)
        return res

    @RaiserRx.raise_if_error
    def index_add(self, namespace: str, index_def: Dict, timeout: timedelta = timedelta(milliseconds=0)) -> None:
        """Adds an index to the specified namespace

        #### Arguments:
            namespace (string): The name of the namespace
            index_def (dict): A dictionary of index definition
            timeout (`datetime.timedelta`): Optional timeout for performing a server-side operation.
                Minimum is 1 millisecond; if set to a lower value, it corresponds to disabling the timeout.
                A value of 0 disables the timeout (default value)

        #### Raises:
            ConnectionError: Raises with an error message when Reindexer instance is not initialized yet
            ApiError: Raises with an error message of API return on non-zero error code

        """

        milliseconds: int = int(timeout / timedelta(milliseconds=1))
        self.err_code, self.err_msg = self.api.index_add(self.rx, namespace, index_def, milliseconds)

    @RaiserRx.raise_if_error
    def index_update(self, namespace: str, index_def: Dict, timeout: timedelta = timedelta(milliseconds=0)) -> None:
        """Updates an index in the specified namespace

        #### Arguments:
            namespace (string): The name of the namespace
            index_def (dict): A dictionary of index definition
            timeout (`datetime.timedelta`): Optional timeout for performing a server-side operation.
                Minimum is 1 millisecond; if set to a lower value, it corresponds to disabling the timeout.
                A value of 0 disables the timeout (default value)

        #### Raises:
            ConnectionError: Raises with an error message when Reindexer instance is not initialized yet
            ApiError: Raises with an error message of API return on non-zero error code

        """

        milliseconds: int = int(timeout / timedelta(milliseconds=1))
        self.err_code, self.err_msg = self.api.index_update(self.rx, namespace, index_def, milliseconds)

    @RaiserRx.raise_if_error
    def index_drop(self, namespace: str, index_name: str, timeout: timedelta = timedelta(milliseconds=0)) -> None:
        """Drops an index from the specified namespace

        #### Arguments:
            namespace (string): The name of the namespace
            index_name (string): A name of an index
            timeout (`datetime.timedelta`): Optional timeout for performing a server-side operation.
                Minimum is 1 millisecond; if set to a lower value, it corresponds to disabling the timeout.
                A value of 0 disables the timeout (default value)

        #### Raises:
            ConnectionError: Raises with an error message when Reindexer instance is not initialized yet
            ApiError: Raises with an error message of API return on non-zero error code

        """

        milliseconds: int = int(timeout / timedelta(milliseconds=1))
        self.err_code, self.err_msg = self.api.index_drop(self.rx, namespace, index_name, milliseconds)

    @RaiserRx.raise_if_error
    def item_insert(self, namespace: str, item_def: Dict, precepts: List[str] = None,
                    timeout: timedelta = timedelta(milliseconds=0)) -> None:
        """Inserts an item with its precepts into the specified namespace

        #### Arguments:
            namespace (string): The name of the namespace
            item_def (dict): A dictionary of item definition
            precepts (:obj:`list` of :obj:`str`): A dictionary of index definition
            timeout (`datetime.timedelta`): Optional timeout for performing a server-side operation.
                Minimum is 1 millisecond; if set to a lower value, it corresponds to disabling the timeout.
                A value of 0 disables the timeout (default value)

        #### Raises:
            ConnectionError: Raises with an error message when Reindexer instance is not initialized yet
            ApiError: Raises with an error message of API return on non-zero error code

        """

        precepts = [] if precepts is None else precepts
        milliseconds: int = int(timeout / timedelta(milliseconds=1))
        self.err_code, self.err_msg = self.api.item_insert(self.rx, namespace, item_def, precepts, milliseconds)

    @RaiserRx.raise_if_error
    def item_update(self, namespace: str, item_def: Dict, precepts: List[str] = None,
                    timeout: timedelta = timedelta(milliseconds=0)) -> None:
        """Updates an item with its precepts in the specified namespace

        #### Arguments:
            namespace (string): The name of the namespace
            item_def (dict): A dictionary of item definition
            precepts (:obj:`list` of :obj:`str`): A dictionary of index definition
            timeout (`datetime.timedelta`): Optional timeout for performing a server-side operation.
                Minimum is 1 millisecond; if set to a lower value, it corresponds to disabling the timeout.
                A value of 0 disables the timeout (default value)

        #### Raises:
            ConnectionError: Raises with an error message when Reindexer instance is not initialized yet
            ApiError: Raises with an error message of API return on non-zero error code

        """

        precepts = [] if precepts is None else precepts
        milliseconds: int = int(timeout / timedelta(milliseconds=1))
        self.err_code, self.err_msg = self.api.item_update(self.rx, namespace, item_def, precepts, milliseconds)

    @RaiserRx.raise_if_error
    def item_upsert(self, namespace: str, item_def: Dict, precepts: List[str] = None,
                    timeout: timedelta = timedelta(milliseconds=0)) -> None:
        """Updates an item with its precepts in the specified namespace. Creates the item if it does not exist

        #### Arguments:
            namespace (string): The name of the namespace
            item_def (dict): A dictionary of item definition
            precepts (:obj:`list` of :obj:`str`): A dictionary of index definition
            timeout (`datetime.timedelta`): Optional timeout for performing a server-side operation.
                Minimum is 1 millisecond; if set to a lower value, it corresponds to disabling the timeout.
                A value of 0 disables the timeout (default value)

        #### Raises:
            ConnectionError: Raises with an error message when Reindexer instance is not initialized yet
            ApiError: Raises with an error message of API return on non-zero error code

        """

        precepts = [] if precepts is None else precepts
        milliseconds: int = int(timeout / timedelta(milliseconds=1))
        self.err_code, self.err_msg = self.api.item_upsert(self.rx, namespace, item_def, precepts, milliseconds)

    @RaiserRx.raise_if_error
    def item_delete(self, namespace: str, item_def: Dict, timeout: timedelta = timedelta(milliseconds=0)) -> None:
        """Deletes an item from the namespace specified

        #### Arguments:
            namespace (string): The name of the namespace
            item_def (dict): A dictionary of item definition
            timeout (`datetime.timedelta`): Optional timeout for performing a server-side operation.
                Minimum is 1 millisecond; if set to a lower value, it corresponds to disabling the timeout.
                A value of 0 disables the timeout (default value)

        #### Raises:
            ConnectionError: Raises with an error message when Reindexer instance is not initialized yet
            ApiError: Raises with an error message of API return on non-zero error code

        """

        milliseconds: int = int(timeout / timedelta(milliseconds=1))
        self.err_code, self.err_msg = self.api.item_delete(self.rx, namespace, item_def, [], milliseconds)

    @RaiserRx.raise_if_error
    def meta_put(self, namespace: str, key: str, value: str, timeout: timedelta = timedelta(milliseconds=0)) -> None:
        """Puts metadata to a storage of Reindexer by key

        #### Arguments:
            namespace (string): The name of the namespace
            key (string): A key in a storage of Reindexer for metadata keeping
            value (string): A metadata for storage
            timeout (`datetime.timedelta`): Optional timeout for performing a server-side operation.
                Minimum is 1 millisecond; if set to a lower value, it corresponds to disabling the timeout.
                A value of 0 disables the timeout (default value)

        #### Raises:
            ConnectionError: Raises with an error message when Reindexer instance is not initialized yet
            ApiError: Raises with an error message of API return on non-zero error code

        """

        milliseconds: int = int(timeout / timedelta(milliseconds=1))
        self.err_code, self.err_msg = self.api.meta_put(self.rx, namespace, key, value, milliseconds)

    @RaiserRx.raise_if_error
    def meta_get(self, namespace: str, key: str, timeout: timedelta = timedelta(milliseconds=0)) -> str:
        """Gets metadata from a storage of Reindexer by key specified

        #### Arguments:
            namespace (string): The name of the namespace
            key (string): A key in a storage of Reindexer where metadata is kept
            timeout (`datetime.timedelta`): Optional timeout for performing a server-side operation.
                Minimum is 1 millisecond; if set to a lower value, it corresponds to disabling the timeout.
                A value of 0 disables the timeout (default value)

        #### Returns:
            string: A metadata value

        #### Raises:
            ConnectionError: Raises with an error message when Reindexer instance is not initialized yet
            ApiError: Raises with an error message of API return on non-zero error code

        """

        milliseconds: int = int(timeout / timedelta(milliseconds=1))
        self.err_code, self.err_msg, res = self.api.meta_get(self.rx, namespace, key, milliseconds)
        return res

    @RaiserRx.raise_if_error
    def meta_delete(self, namespace: str, key: str, timeout: timedelta = timedelta(milliseconds=0)) -> None:
        """Deletes metadata from a storage of Reindexer by key specified

        #### Arguments:
            namespace (string): The name of the namespace
            key (string): A key in a storage of Reindexer where metadata is kept
            timeout (`datetime.timedelta`): Optional timeout for performing a server-side operation.
                Minimum is 1 millisecond; if set to a lower value, it corresponds to disabling the timeout.
                A value of 0 disables the timeout (default value)

        #### Raises:
            ConnectionError: Raises with an error message when Reindexer instance is not initialized yet
            ApiError: Raises with an error message of API return on non-zero error code

        """

        milliseconds: int = int(timeout / timedelta(milliseconds=1))
        self.err_code, self.err_msg = self.api.meta_delete(self.rx, namespace, key, milliseconds)

    @RaiserRx.raise_if_error
    def meta_enum(self, namespace: str, timeout: timedelta = timedelta(milliseconds=0)) -> List[str]:
        """Gets a list of metadata keys from a storage of Reindexer

        #### Arguments:
            namespace (string): The name of the namespace
            timeout (`datetime.timedelta`): Optional timeout for performing a server-side operation.
                Minimum is 1 millisecond; if set to a lower value, it corresponds to disabling the timeout.
                A value of 0 disables the timeout (default value)

        #### Returns:
            (:obj:`list` of :obj:`str`): A list of all metadata keys

        #### Raises:
            ConnectionError: Raises with an error message when Reindexer instance is not initialized yet
            ApiError: Raises with an error message of API return on non-zero error code

        """

        milliseconds: int = int(timeout / timedelta(milliseconds=1))
        self.err_code, self.err_msg, res = self.api.meta_enum(self.rx, namespace, milliseconds)
        return res

    @RaiserRx.raise_if_error
    def exec_sql(self, query: str, timeout: timedelta = timedelta(milliseconds=0)) -> QueryResults:
        """Executes an SQL query and returns query results

        #### Arguments:
            query (string): An SQL query
            timeout (`datetime.timedelta`): Optional timeout for performing a server-side operation.
                Minimum is 1 millisecond; if set to a lower value, it corresponds to disabling the timeout.
                A value of 0 disables the timeout (default value)

        #### Returns:
            (:obj:`QueryResults`): A QueryResults iterator

        #### Raises:
            ConnectionError: Raises with an error message when Reindexer instance is not initialized yet
            ApiError: Raises with an error message of API return on non-zero error code

        """

        milliseconds: int = int(timeout / timedelta(milliseconds=1))
        self.err_code, self.err_msg, wrapper_ptr, iter_count, total_count = self.api.exec_sql(self.rx, query,
                                                                                              milliseconds)
        return QueryResults(self.api, wrapper_ptr, iter_count, total_count)

    @RaiserRx.raise_if_error
    def new_transaction(self, namespace: str, timeout: timedelta = timedelta(milliseconds=0)) -> Transaction:
        """Starts a new transaction and return the transaction object to processing.
            Warning: once a timeout is set, it will apply to all subsequent steps in the transaction

        #### Arguments:
            namespace (string): The name of the namespace
            timeout (`datetime.timedelta`): Optional timeout for performing a server-side operation.
                Minimum is 1 millisecond; if set to a lower value, it corresponds to disabling the timeout.
                A value of 0 disables the timeout (default value)

        #### Returns:
            (:obj:`Transaction`): A new transaction

        #### Raises:
            ConnectionError: Raises with an error message when Reindexer instance is not initialized yet
            ApiError: Raises with an error message of API return on non-zero error code

        """

        milliseconds: int = int(timeout / timedelta(milliseconds=1))
        self.err_code, self.err_msg, transaction_wrapper_ptr = self.api.new_transaction(self.rx, namespace,
                                                                                        milliseconds)
        transaction = Transaction(self, transaction_wrapper_ptr)
        with self._tx_lock:
            self._tx_ptrs.add(transaction_wrapper_ptr)
        return transaction

    @RaiserRx.raise_if_error
    def new_query(self, namespace: str) -> Query:
        """Creates a new query and return the query object to processing

        #### Arguments:
            namespace (string): The name of the namespace

        #### Returns:
            (:obj:`Query`): A new query

        #### Raises:
            ConnectionError: Raises with an error message when Reindexer instance is not initialized yet

        """

        self.err_code, self.err_msg, query_wrapper_ptr = self.api.create_query(self.rx, namespace)
        query = Query(self, query_wrapper_ptr)
        with self._query_lock:
            self._query_ptrs.add(query_wrapper_ptr)
        return query



from datetime import timedelta

from hamcrest import *


class TestCrudMetadata:
    def test_initial_namespace_has_no_metadata(self, db, namespace):
        # Given("Create namespace")
        # When ("Get list of metadata keys")
        meta_list = db.meta.enumerate(namespace)
        # Then ("Check that list of metadata is empty")
        assert_that(meta_list, empty(), "Metadata is not empty")

    def test_create_metadata(self, db, namespace):
        # Given("Create namespace")
        # When ("Put metadata to namespace")
        key, value = 'key', 'value'
        db.meta.put(namespace, key, value)
        # Then ("Check that metadata was added")
        meta_list = db.meta.enumerate(namespace)
        assert_that(meta_list, has_length(1), "Metadata is not created")
        assert_that(meta_list, has_item(key), "Metadata is not created")

    def test_get_metadata_by_key(self, db, namespace, metadata):
        # Given ("Put metadata to namespace")
        meta_key, meta_value = metadata
        # When ("Get metadata by key")
        value = db.meta.get(namespace, meta_key)
        # Then ("Check that metadata was added")
        assert_that(value, equal_to(meta_value), "Can't get meta value by key")

    def test_get_metadata_keys(self, db, namespace, metadata):
        # Given ("Put metadata to namespace")
        meta_key, meta_value = metadata
        # When ("Get list of metadata keys")
        meta_list = db.meta.enumerate(namespace)
        # Then ("Check that metadata was added")
        assert_that(meta_list, has_item(meta_key), "Can't get list of meta keys")

    def test_update_metadata(self, db, namespace, metadata):
        # Given ("Put metadata to namespace")
        meta_key, meta_value = metadata
        # When ("Update metadata")
        new_meta_value = 'new_value'
        db.meta.put(namespace, meta_key, new_meta_value)
        # Then ("Check that metadata was updated")
        updated_value = db.meta.get(namespace, meta_key)
        assert_that(updated_value, equal_to(new_meta_value), "Can't update metadata")

    def test_delete_metadata(self, db, namespace, metadata):
        # Given ("Put metadata to namespace")
        meta_key, meta_value = metadata
        # When ("Delete metadata")
        db.meta.delete(namespace, meta_key)
        # Then ("Check that metadata was removed")
        read_value = db.meta.get(namespace, meta_key)
        assert_that(read_value, equal_to(''), "Can't delete metadata")
        # When ("Get list of metadata keys")
        meta_list = db.meta.enumerate(namespace)
        # Then ("Check that list of metadata is empty")
        assert_that(meta_list, empty(), "Metadata is not empty")

    def test_metadata_put_and_delete_timeouts(self, db, namespace):
        # Given("Create namespace")
        # When ("Put metadata to namespace with big timeout")
        key, value = "key", "value"
        db.meta.put(namespace, key, value, timeout=timedelta(milliseconds=1000))
        # Then ("Check that metadata was added")
        meta_list = db.meta.enumerate(namespace, timeout=timedelta(milliseconds=1000))
        assert_that(meta_list, has_length(1), "Metadata is not created")
        assert_that(meta_list, has_item(key), "Metadata is not created")
        # When ("Delete metadata with big timeout")
        db.meta.delete(namespace, key, timeout=timedelta(milliseconds=1000))
        # Then ("Check that metadata was removed")
        read_value = db.meta.get(namespace, key)
        assert_that(read_value, equal_to(''), "Can't delete metadata")
        # When ("Get list of metadata keys")
        meta_list = db.meta.enumerate(namespace)
        # Then ("Check that list of metadata is empty")
        assert_that(meta_list, empty(), "Metadata is not empty")



import datetime
import os
import shlex
import subprocess

import requests
from requests.adapters import HTTPAdapter, Retry

from tests.helpers.log_helper import log_fixture


session = requests.Session()
Retry.DEFAULT_BACKOFF_MAX = 0.03
retry = Retry(total=100, backoff_factor=0.03)
adapter = HTTPAdapter(max_retries=retry)
session.mount("http://", adapter)


class ReindexerServer:

    def __init__(self, rx_bin_path="", http_port=9088, rpc_port=6534, storage="/tmp/reindex_test",
                 auth=False, user=None, password=None):
        self.rx_bin_path = rx_bin_path
        self.http_port = http_port
        self.rpc_port = rpc_port
        self.httpaddr = f"127.0.0.1:{http_port}"
        self.rpcaddr = f"127.0.0.1:{rpc_port}"

        self.storage = storage

        self.auth = auth
        self.user = user
        self.password = password

        self.proc = None

        self.__cmd = f"{os.path.join(self.rx_bin_path, 'reindexer_server')} --db {self.storage} --httpaddr={self.httpaddr} --rpcaddr={self.rpcaddr}"
        if self.auth:
            self.__cmd += " --security"

    @staticmethod
    def _server_log_file(module):
        log_folder = os.path.join(os.path.dirname(os.path.abspath(__file__)), "../logs/server")
        if not os.path.exists(log_folder):
            os.makedirs(log_folder, exist_ok=True)
        if module and "/" in module:
            file_name = module.split("/")
            module = file_name[-1]
            parent = f"{file_name[-2]}_"
        elif module:
            parent = ""
        else:
            module = ""
            parent = ""
        time_suffix = datetime.datetime.now().strftime('%d%b-%H_%M_%S.%f')
        log_path = os.path.join(log_folder, f"{parent}{module}_{time_suffix}.txt")
        log_file = open(log_path, "w+", buffering=1)
        return log_file

    def run(self, *args, **kwargs):
        command = shlex.split(self.__cmd) + list(args)
        log_file = self._server_log_file(kwargs.get("module"))
        self.proc = subprocess.Popen(command, stdout=log_file, stderr=log_file, cwd=self.rx_bin_path or None)
        self.is_running()
        log_fixture.info(f"Reindexer server started at {self.http_port} http port")

    def is_running(self):
        try:
            url = f"http://{self.httpaddr}/api/v1/check"
            auth = (self.user, self.password) if self.auth else None
            session.get(url=url, auth=auth)
            return True
        except (ConnectionError, requests.exceptions.ConnectionError, RecursionError, ChildProcessError,
                AssertionError):
            self.terminate()
            if not hasattr(self.proc, "returncode"):
                raise ChildProcessError("Server is not running")
            else:
                raise ChildProcessError(f"Server status code is {self.proc.returncode}")

    def terminate(self):
        try:
            self.proc.terminate()
            self.proc.wait(timeout=1)
        except subprocess.TimeoutExpired:
            self.proc.kill()



from datetime import timedelta

from hamcrest import *

from pyreindexer.exceptions import ApiError


class TestCrudNamespace:

    def test_create_ns(self, db):
        # Given("Create namespace in empty database")
        namespace_name = 'test_ns1'
        db.namespace.open(namespace_name, timeout=timedelta(seconds=1))
        # When ("Get namespaces list in created database")
        namespace_list = db.namespace.enumerate(timeout=timedelta(milliseconds=1000))
        # Then ("Check that database contains created namespace")
        assert_that(namespace_list, has_item(has_entries(name=namespace_name)),
                    "Namespace wasn't created")
        db.namespace.drop(namespace_name, timeout=timedelta(milliseconds=1000))

    def test_drop_ns(self, db):
        # Given("Create namespace in empty database")
        namespace_name = 'test_ns2'
        db.namespace.open(namespace_name)
        # When ("Delete namespace")
        db.namespace.drop(namespace_name)
        # Then ("Check that namespace was deleted")
        namespace_list = db.namespace.enumerate()
        assert_that(namespace_list, not_(has_item(has_entries(name=namespace_name))),
                    "Namespace wasn't deleted")

    def test_cannot_delete_ns_not_created(self, db):
        # Given ("Created empty database")
        # Then ("Check that we cannot delete namespace that does not exist")
        namespace_name = 'test_ns'
        assert_that(calling(db.namespace.drop).with_args(namespace_name),
                    raises(ApiError, pattern=f"Namespace '{namespace_name}' does not exist"))



import copy
from datetime import timedelta
from typing import Final

import pytest
from hamcrest import *

from pyreindexer.exceptions import ApiError
from tests.helpers.base_helper import get_ns_description
from tests.test_data.constants import (index_definition, updated_index_definition, vector_index_bf, vector_index_hnsw,
                                       vector_index_ivf)


class TestCrudIndexes:
    def test_initial_namespace_has_no_indexes(self, db, namespace):
        # Given("Create namespace")
        # When ("Get namespace information")
        ns_entry = get_ns_description(db, namespace)
        # Then ("Check that list of indexes in namespace is empty")
        assert_that(ns_entry, has_item(has_entry("indexes", empty())), "Index is not empty")

    def test_create_index(self, db, namespace):
        # Given("Create namespace")
        # When ("Add index")
        db.index.create(namespace, index_definition, timeout=timedelta(milliseconds=1000))
        # Then ("Check that index is added")
        ns_entry = get_ns_description(db, namespace)
        assert_that(ns_entry, has_item(has_entry("indexes", has_item(index_definition))),
                    "Index wasn't created")

    @pytest.mark.parametrize("vector_index", [vector_index_bf, vector_index_hnsw, vector_index_ivf])
    def test_create_vector_index(self, db, namespace, vector_index):
        # Given("Create namespace")
        # When ("Add index")
        db.index.create(namespace, vector_index)
        # Then ("Check that index is added")
        ns_entry = get_ns_description(db, namespace)
        assert_that(ns_entry, has_item(has_entry("indexes", has_item(vector_index))),
                    "Index wasn't created")

    def test_update_index(self, db, namespace, index):
        # Given("Create namespace with index")
        # When ("Update index")
        db.index.update(namespace, updated_index_definition)
        # Then ("Check that index is updated")
        ns_entry = get_ns_description(db, namespace)
        assert_that(ns_entry, has_item(has_entry("indexes", has_item(updated_index_definition))),
                    "Index wasn't updated")

    def test_delete_index(self, db, namespace):
        # Given("Create namespace with index")
        db.index.create(namespace, index_definition)
        # When ("Delete index")
        db.index.drop(namespace, 'id')
        # Then ("Check that index is deleted")
        ns_entry = get_ns_description(db, namespace)
        assert_that(ns_entry, has_item(has_entry("indexes", empty())), "Index wasn't deleted")

    def test_cannot_add_index_with_same_name(self, db, namespace):
        # Given("Create namespace")
        # When ("Add index")
        db.index.create(namespace, index_definition)
        # Then ("Check that we can't add index with the same name")
        assert_that(calling(db.index.create).with_args(namespace, updated_index_definition),
                    raises(ApiError, pattern="Index '.*' already exists with different settings"),
                    "Index with existing name was created")

    def test_cannot_update_not_existing_index_in_namespace(self, db, namespace):
        # Given ("Create namespace")
        # When ("Update index")
        # Then ("Check that we can't update index that was not created")
        assert_that(calling(db.index.update).with_args(namespace, index_definition),
                    raises(ApiError, pattern=f"Index 'id' not found in '{namespace}'"),
                    "Not existing index was updated")

    def test_cannot_delete_not_existing_index_in_namespace(self, db, namespace):
        # Given ("Create namespace")
        # When ("Delete index")
        # Then ("Check that we can't delete index that was not created")
        index_name = 'id'
        assert_that(calling(db.index.drop).with_args(namespace, index_name),
                    raises(ApiError, pattern=f"Cannot remove index '{index_name}': doesn't exist"),
                    "Not existing index was deleted")

    def test_index_update_timeout(self, db, namespace):
        # Given("Create index")
        db.index.create(namespace, index_definition, timeout=timedelta(milliseconds=1000))
        # When ("Update index with big timeout")
        db.index.update(namespace, updated_index_definition, timeout=timedelta(milliseconds=1000))
        # Then ("Check that index is updated")
        ns_entry = get_ns_description(db, namespace)
        assert_that(ns_entry, has_item(has_entry("indexes", has_item(updated_index_definition))),
                    "Index wasn't updated")

    def test_invalid_index_embedder_config(self, db, namespace, index):
        # Given ("Create float vector index")
        dimension: Final[int] = 5
        index = copy.deepcopy(vector_index_hnsw)
        index["config"]["dimension"] = dimension
        index["config"]["embedding"] = {
            "query_embedder": {"URL": "abc"}
        }
        err_msg = "Configuration 'embedding:query_embedder' contain field 'URL' with unexpected value: 'abc'"
        assert_that(calling(db.index.create).with_args(namespace, index), raises(ApiError, pattern=err_msg))



from datetime import timedelta
from math import nan, inf

import pytest
from hamcrest import *

from tests.helpers.base_helper import get_ns_items, get_ns_vect_items, random_vector
from tests.helpers.matchers import close_to_dict
from tests.test_data.constants import item_definition, vector_index_bf, vector_index_hnsw, vector_index_ivf


class TestCrudItems:
    def test_initial_namespace_has_no_items(self, db, namespace, index):
        # Given("Create namespace with index")
        # When ("Get namespace information")
        select_result = get_ns_items(db, namespace)
        # Then ("Check that list of items in namespace is empty")
        assert_that(select_result, empty(), "Item list is not empty")

    def test_create_item_insert(self, db, namespace, index):
        # Given("Create namespace with index")
        # When ("Insert item into namespace")
        db.item.insert(namespace, item_definition)
        # Then ("Check that item is added")
        select_result = get_ns_items(db, namespace)
        assert_that(select_result, has_length(1), "Item wasn't created")
        assert_that(select_result, has_item(item_definition), "Item wasn't created")

    @pytest.mark.parametrize("value", [-1, -2.33])
    def test_create_item_insert_negative(self, db, namespace, index, value):
        # Given("Create namespace with index")
        # When ("Insert item into namespace")
        item_empty = {"id": -1, "field": value, "empty_field": None, "nan_field": nan, "inf_field": inf,
                      "ninf_field": -inf}
        db.item.insert(namespace, item_empty)
        # Then ("Check that item is added")
        select_result = get_ns_items(db, namespace)
        assert_that(select_result, has_length(1), "Item wasn't created")
        item_empty["nan_field"] = None
        item_empty["inf_field"] = None
        item_empty["ninf_field"] = None
        assert_that(select_result, has_item(item_empty), "Item wasn't created")

    @pytest.mark.parametrize("vector_index", [vector_index_bf, vector_index_hnsw, vector_index_ivf])
    def test_create_items_with_vector_index(self, db, namespace, index, vector_index):
        # Given("Create namespace")
        # When ("Add index")
        db.index.create(namespace, vector_index)
        # When ("Insert item")
        dimension = vector_index["config"]["dimension"]
        item = {"id": 0, "vec": random_vector(dimension)}
        db.item.insert(namespace, item)
        # Then ("Check that item is added")
        select_result = get_ns_vect_items(db, namespace)
        assert_that(select_result, has_length(1), "Item wasn't created")
        assert_that(select_result[0], close_to_dict(item))

    def test_create_item_insert_with_precepts(self, db, namespace, index):
        # Given("Create namespace with index")
        # When ("Insert items into namespace")
        number_items = 5
        for _ in range(number_items):
            db.item.insert(namespace, {"id": 100, "field": "value"}, ["id=serial()"])
        # Then ("Check that item is added")
        select_result = get_ns_items(db, namespace)
        assert_that(select_result, has_length(number_items), "Items wasn't created")
        for i in range(number_items):
            assert_that(select_result[i], equal_to({'id': i + 1, "field": "value"}), "Items wasn't created")

    def test_create_item_upsert(self, db, namespace, index):
        # Given("Create namespace with index")
        # When ("Upsert item into namespace")
        db.item.upsert(namespace, item_definition)
        # Then ("Check that item is added")
        select_result = get_ns_items(db, namespace)
        assert_that(select_result, has_length(1), "Item wasn't created")
        assert_that(select_result, has_item(item_definition), "Item wasn't created")

    def test_update_item_upsert(self, db, namespace, index, item):
        # Given("Create namespace with item")
        # When ("Upsert item")
        item_definition_updated = {'id': 100, 'val': "new_value"}
        db.item.upsert(namespace, item_definition_updated)
        # Then ("Check that item is updated")
        select_result = get_ns_items(db, namespace)
        assert_that(select_result, has_length(1), "Item wasn't updated")
        assert_that(select_result, has_item(item_definition_updated), "Item wasn't updated")

    def test_update_item_update(self, db, namespace, index, item):
        # Given("Create namespace with item")
        # When ("Update item")
        item_definition_updated = {'id': 100, 'val': "new_value"}
        db.item.update(namespace, item_definition_updated)
        # Then ("Check that item is updated")
        select_result = get_ns_items(db, namespace)
        assert_that(select_result, has_length(1), "Item wasn't updated")
        assert_that(select_result, has_item(item_definition_updated), "Item wasn't updated")

    def test_delete_item(self, db, namespace, index, item):
        # Given("Create namespace with item")
        # When ("Delete item")
        db.item.delete(namespace, item_definition)
        # Then ("Check that item is deleted")
        select_result = get_ns_items(db, namespace)
        assert_that(select_result, empty(), "Item wasn't deleted")

    def test_item_timeouts(self, db, namespace, index):
        # Given("Create namespace with index")
        # When ("Insert item into namespace with big timeout")
        db.item.insert(namespace, item_definition, timeout=timedelta(milliseconds=1000))
        # Then ("Check that item is added")
        select_result = get_ns_items(db, namespace)
        assert_that(select_result, has_length(1), "Item wasn't created")
        assert_that(select_result, has_item(item_definition), "Item wasn't created")
        # When ("Update item with big timeout")
        item_definition_updated = {'id': 100, 'val': "new_value"}
        db.item.update(namespace, item_definition_updated, timeout=timedelta(milliseconds=1000))
        # Then ("Check that item is updated")
        select_result = get_ns_items(db, namespace)
        assert_that(select_result, has_length(1), "Item wasn't updated")
        assert_that(select_result, has_item(item_definition_updated), "Item wasn't updated")
        # When ("Delete item with big timeout")
        db.item.delete(namespace, item_definition_updated, timeout=timedelta(milliseconds=1000))
        # Then ("Check that item is deleted")
        select_result = get_ns_items(db, namespace)
        assert_that(select_result, empty(), "Item wasn't deleted")



# -*- coding: utf-8 -*-
import logging
import os
import sys
from datetime import *


class OneLineExceptionFormatter(logging.Formatter):
    """ One line log output
    https://docs.python.org/2/howto/logging-cookbook.html#logging-cookbook
    """

    def formatException(self, exc_info):
        result = super(OneLineExceptionFormatter, self).formatException(exc_info)
        return repr(result)

    def format(self, record):
        result = super(OneLineExceptionFormatter, self).format(record)
        if record.exc_text:
            result = result.replace('\n', '') + '|'
        return result


# Create logger
log_api = logging.getLogger('API')
log_fixture = logging.getLogger('FIXTURE')
log_error = logging.getLogger('ERROR')

log_api.setLevel(logging.INFO)
log_fixture.setLevel(logging.INFO)
log_error.setLevel(logging.ERROR)

# Log format
formatter = OneLineExceptionFormatter('%(levelname)-5s [%(asctime)s] [%(name)s]: %(message)s')

# Log to console
console_handler = logging.StreamHandler(sys.stdout)
console_handler.setLevel(logging.INFO)
console_handler.setFormatter(formatter)
log_api.addHandler(console_handler)
log_fixture.addHandler(console_handler)

# Save log to file
log_folder = os.path.join(os.path.dirname(os.path.abspath(__file__)), '../logs/')
if not os.path.exists(log_folder):
    os.makedirs(log_folder)
filename = f"test_run_{datetime.now().strftime('%d%b-%H_%M_%S')}.txt"
log_filepath = os.path.join(os.path.dirname(log_folder), filename)

file_handler = logging.FileHandler(log_filepath)
file_handler.setLevel(logging.INFO)
file_handler.setFormatter(formatter)
log_api.addHandler(file_handler)
log_fixture.addHandler(file_handler)



import os
import sys

from setuptools import Extension, setup
from setuptools.command.build_ext import build_ext as build_ext_orig


if sys.version_info < (3, 8):
    raise RuntimeError('Require Python 3.8 or greater')

PACKAGE_NAME = 'pyreindexer'


def _c2(*names):
    return ' :: '.join(names)


class CMakeExtension(Extension):
    def __init__(self, name):
        super().__init__(name, sources=[])


class BuildExt(build_ext_orig):
    def run(self):
        for ext in self.extensions:
            self.build_cmake(ext)

    def build_cmake(self, ext):
        os.makedirs(os.path.abspath(self.build_temp), exist_ok=True)

        cwd = os.path.abspath('')
        extension_dir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
        self.spawn(['cmake',
                    os.path.join(cwd, PACKAGE_NAME),
                    f'-DCMAKE_LIBRARY_OUTPUT_DIRECTORY={extension_dir}',
                    '-DCMAKE_BUILD_TYPE=Release',
                    '-DCMAKE_CXX_STANDARD=20',
                    f'-DCMAKE_OSX_DEPLOYMENT_TARGET={os.getenv("MACOSX_DEPLOYMENT_TARGET", "14")}'])

        if not self.dry_run:
            self.spawn(['cmake', '--build', '.'])
        os.chdir(cwd)


with open('README.md', 'r', encoding='utf-8') as file:
    LONG_DESCRIPTION = file.read()

setup(name=PACKAGE_NAME,
      version='0.5.90003',
      description='A connector that allows to interact with Reindexer. Reindexer static library or reindexer-dev package must be installed',
      author='Igor Tulmentyev',
      author_email='contactus@reindexer.io',
      maintainer='Reindexer Team',
      maintainer_email='contactus@reindexer.io',
      url='https://github.com/Restream/reindexer',
      download_url='https://github.com/Restream/reindexer-py',
      project_urls={
          'Documentation': 'https://reindexer.io/',
          # 'Releases': 'https://github.com/Restream/reindexer-py/releases',
          # 'Tracker': 'https://github.com/Restream/reindexer-py/issues',
          'Telegram chat': 'https://t.me/reindexer',
      },
      long_description=LONG_DESCRIPTION,
      long_description_content_type="text/markdown",
      license='Apache License 2.0',
      packages=[PACKAGE_NAME],
      ext_modules=[CMakeExtension('rawpyreindexer')],
      cmdclass={'build_ext': BuildExt},
      keywords=['reindexer', 'reindexer-py', 'in-memory-database', 'database', 'python', 'connector'],
      package_data={'pyreindexer': [
          'CMakeLists.txt',
          'lib/**/*',
          'example/main.py',
          'tests/**/*.py'
      ]},
      python_requires='>=3.8',
      install_requires=['PyHamcrest==2.0.2', 'pytest>=6.2.5', 'requests>=2.26.0'],
      classifiers=[
          _c2('Development Status', '4 - Beta'),
          _c2('Environment', 'Console'),
          _c2('Intended Audience', 'End Users/Desktop'),
          _c2('Intended Audience', 'Developers'),
          _c2('Natural Language', 'Russian'),
          _c2('Operating System', 'MacOS'),
          _c2('Operating System', 'POSIX', 'Linux'),
          _c2('Programming Language', 'Python'),
          _c2('Programming Language', 'Python', '3', 'Only'),
          _c2('Programming Language', 'Python', '3.8'),
          _c2('Programming Language', 'Python', '3.9'),
          _c2('Programming Language', 'Python', '3.10'),
          _c2('Programming Language', 'Python', '3.11'),
          _c2('Programming Language', 'Python', '3.12'),
          _c2('Programming Language', 'Python', '3.13'),
          _c2('Programming Language', 'Python', 'Implementation'),
          _c2('Programming Language', 'Python', 'Implementation', 'CPython'),
          _c2('Programming Language', 'Python', 'Implementation', 'PyPy'),
          _c2('Topic', 'Database'),
          _c2('Topic', 'Database', 'Database Engines/Servers'),
          _c2('Topic', 'Software Development'),
          _c2('Topic', 'Software Development', 'Libraries'),
          _c2('Topic', 'Software Development', 'Libraries', 'Python Modules'),
          _c2('Topic', 'Software Development', 'Libraries', 'Application Frameworks'),
      ],
      platforms=['Ubuntu', 'Debian', 'ALT Linux', 'RED OS', 'Astra Linux', 'MacOS'],
      )



from hamcrest import *

from tests.test_data.constants import special_namespaces, special_namespaces_cluster


class TestCrudDb:
    def test_create_db(self, db):
        # When ("Get namespaces list in created database")
        namespaces_list = db.namespace.enumerate()
        # Then ("Check that database contains only special namespaces")
        if len(namespaces_list) == len(special_namespaces_cluster):
            expected_namespaces = special_namespaces_cluster  # v4
        else:
            expected_namespaces = special_namespaces  # v3
        assert_that(namespaces_list, has_length(len(expected_namespaces)),
                    "Database doesn't contain special namespaces")
        for namespace in expected_namespaces:
            assert_that(namespaces_list, has_item(has_entries(name=namespace["name"])),
                        "Database doesn't contain special namespaces")



class Point:
    """An object representing the context of a Reindexer 2D point

    #### Attributes:
        x (float): x coordinate of the point
        y (float): y coordinate of the point
    """

    def __init__(self, x: float, y: float):
        """Constructs a new Reindexer query object

        #### Arguments:
            x (float): x coordinate of the point
            y (float): y coordinate of the point

        """

        self.x: float = x
        self.y: float = y
